import warnings
import numpy as np
import pandas as pd
from typing import Any

from mirp._data_import.dicom_file import ImageDicomFile
from mirp._data_import.generic_file import MaskFile
from mirp._data_import.generic_file_stack import ImageFileStack, MaskFileStack
from mirp._data_import.utilities import get_pydicom_meta_tag


class ImageDicomFileStack(ImageFileStack):

    def __init__(
            self,
            image_file_objects: list[ImageDicomFile],
            **kwargs
    ):
        super().__init__(image_file_objects, **kwargs)

        self.series_instance_uid: None | str = None
        self.frame_of_reference_uid: None | str = None

        # Placeholder for sop_instance_uid of objects contained in the stack.
        self.sop_instance_uid: None | list[str] = None

        # Add type hint.
        self.image_file_objects: list[ImageDicomFile] = self.image_file_objects

    def associate_with_mask(
            self,
            mask_list,
            association_strategy=None):
        if mask_list is None or len(mask_list) == 0 or association_strategy is None:
            return None

        # Match on frame of reference UID:
        if "frame_of_reference" in association_strategy and self.frame_of_reference_uid is not None:
            matching_mask_list = [
                mask_file for mask_file in mask_list
                if self.frame_of_reference_uid == mask_file.frame_of_reference_uid
            ]

            if len(matching_mask_list) > 0:
                self.associated_masks = matching_mask_list
                return

        return super().associate_with_mask(mask_list=mask_list, association_strategy=association_strategy)

    def complete(self, remove_metadata=False, force=False):
        """
        Fills out missing attributes in an image stack. Image parameters in DICOM stacks, by design,
        are fully determined by the origin of all slices in the stack. This method then sorts the image file objects
        by origin, and uses their relative positions to determine slice spacing and the orientation vector.
        :param remove_metadata: Whether metadata (DICOM headers) should be removed after completing information.
        :param force: Whether attributes are forced to update or not.
        :return: nothing, attributes are updated in place.
        """

        # Load metadata of every slice.
        self.load_metadata(limited=True)

        self._complete_modality()
        self._complete_sample_name()

        # Set image orientation.
        # First read orientation (Xx, Xy, Xz, Yx, Yy, Yz) from metadata.
        image_orientation: list[float] = get_pydicom_meta_tag(
            dcm_seq=self.image_file_objects[0].image_metadata,
            tag=(0x0020, 0x0037),
            tag_type="mult_float"
        )
        image_orientation += list(np.cross(image_orientation[0:3], image_orientation[3:6]))

        # Revert to z, y, x order and reshape to matrix.
        if self.image_orientation is None:
            self.image_orientation = np.reshape(image_orientation[::-1], [3, 3], order="F")

        # Set image origin.
        # The image origin is the origin of the first slice, i.e. the slice for which any negative shift in voxel
        # space would increase distance from all other slice origins. Thus, one of the two outermost slices is the
        # first slice.
        position_table = self._get_origin_position_table()

        # First, find the origins of the top and bottom layers:
        outer_origin_0 = position_table.head(1)
        outer_origin_0 = np.array([
            outer_origin_0.position_z.values[0],
            outer_origin_0.position_y.values[0],
            outer_origin_0.position_x.values[0]
        ])

        outer_origin_1 = position_table.tail(1)
        outer_origin_1 = np.array([
            outer_origin_1.position_z.values[0],
            outer_origin_1.position_y.values[0],
            outer_origin_1.position_x.values[0]
        ])

        # Determine distance between bottom and top layers.
        layer_distance = np.sqrt(np.sum(np.power(outer_origin_1 - outer_origin_0, 2.0)))

        # Determine distance between the top layer and a layer outside below the volume, assuming that the bottom
        # layer has the origin at outer_origin_0.
        out_of_volume_layer = self.to_world_coordinates(
            np.array([-1.0, 0.0, 0.0]),
            origin=outer_origin_0,
            spacing=(1.0, 1.0, 1.0)
        )
        oov_layer_distance = np.sqrt(np.sum(np.power(outer_origin_1 - out_of_volume_layer, 2.0)))

        # If the distance between the origins of the top layer and the out-of-volume layer is indeed larger
        # than the distance between top and bottom layer, the assumption that the origin is located at the bottom
        # layer is correct. Otherwise, the origin lies at the top layer instead.
        if self.image_origin is None:
            if oov_layer_distance > layer_distance:
                self.image_origin = tuple(outer_origin_0)
            else:
                self.image_origin = tuple(outer_origin_1)

        # Set image spacing
        # Compute the distance between the origins of the slices. This is the slice spacing.
        image_slice_spacing = np.sqrt(
            np.power(np.diff(position_table.position_x.values), 2.0) +
            np.power(np.diff(position_table.position_y.values), 2.0) +
            np.power(np.diff(position_table.position_z.values), 2.0))

        # Find the smallest slice spacing.
        min_slice_spacing = np.min(image_slice_spacing)

        # Find how much other slices differ.
        image_slice_spacing_multiplier = image_slice_spacing / min_slice_spacing

        if np.any(image_slice_spacing_multiplier > 1.2):
            warnings.warn(
                f"Inconsistent distance between slice origins of subsequent slices: {np.unique(image_slice_spacing)}. "
                f"Slices cannot be aligned correctly. This is likely due to missing slices. "
                f"MIRP will attempt to interpolate the missing slices and their ROI masks for volumetric analysis. "
                f"[{self.describe_self()}]",
                UserWarning)

        # Determine image slice spacing.
        image_slice_spacing = np.around(np.mean(image_slice_spacing[image_slice_spacing_multiplier <= 1.2]), 5)

        # Warn the user if there is a mismatch between slice thickness and the actual slice spacing.
        image_slice_thickness = get_pydicom_meta_tag(
            dcm_seq=self.image_file_objects[0].image_metadata,
            tag=(0x0018, 0x0050),
            tag_type="float")

        if not np.around(image_slice_thickness - image_slice_spacing, decimals=3) == 0.0:
            warnings.warn(
                f"Mismatch between slice thickness ({image_slice_thickness}) and actual slice spacing "
                f"({image_slice_spacing}). The actual slice spacing will be used. [{self.describe_self()}]",
                UserWarning)

        image_pixel_spacing = get_pydicom_meta_tag(
            dcm_seq=self.image_file_objects[0].image_metadata,
            tag=(0x0028, 0x0030),
            tag_type="mult_float")

        # Set image spacing.
        if self.image_spacing is None:
            self.image_spacing = tuple([image_slice_spacing, image_pixel_spacing[1], image_pixel_spacing[0]])

        # Sort image objects in order of ascending position in voxel space.
        layer_positions = [
            self.to_voxel_coordinates(
                x=np.array(get_pydicom_meta_tag(
                    dcm_seq=image_object.image_metadata,
                    tag=(0x0020, 0x0032),
                    tag_type="mult_float",
                    default=np.array([0.0, 0.0, 0.0])
                )[::-1])
            )[0]
            for image_object in self.image_file_objects
        ]

        # Determine how image objects should be sorted according to increasing layer positions.
        new_order = np.argsort(layer_positions)
        self.image_file_objects = [self.image_file_objects[ii] for ii in new_order]

        # Set image dimensions. First, find the number of rows (y) and columns (x) in the data set.
        n_x = get_pydicom_meta_tag(
            dcm_seq=self.image_file_objects[0].image_metadata,
            tag=(0x0028, 0x0011),
            tag_type="int")
        n_y = get_pydicom_meta_tag(
            dcm_seq=self.image_file_objects[0].image_metadata,
            tag=(0x0028, 0x0010),
            tag_type="int")

        if self.image_dimension is None:
            self.image_dimension = tuple([len(position_table), n_y, n_x])

        if self.sop_instance_uid is None:
            self.sop_instance_uid = [image.sop_instance_uid for image in self.image_file_objects]
        if self.frame_of_reference_uid is None:
            self.frame_of_reference_uid = self.image_file_objects[0].frame_of_reference_uid
        if self.series_instance_uid is None:
            self.series_instance_uid = self.image_file_objects[0].series_instance_uid

        # Check if the complete data passes verification.
        self.check(raise_error=True, remove_metadata=False)

        if remove_metadata:
            self.remove_metadata()

    def _get_origin_position_table(self) -> pd.DataFrame:
        # Placeholders for slice positions.
        image_position_z = [0.0] * len(self.image_file_objects)
        image_position_y = [0.0] * len(self.image_file_objects)
        image_position_x = [0.0] * len(self.image_file_objects)

        for ii, image_object in enumerate(self.image_file_objects):
            slice_origin = get_pydicom_meta_tag(
                dcm_seq=image_object.image_metadata,
                tag=(0x0020, 0x0032),
                tag_type="mult_float",
                default=np.array([0.0, 0.0, 0.0]))[::-1]

            image_position_z[ii] = slice_origin[0]
            image_position_y[ii] = slice_origin[1]
            image_position_x[ii] = slice_origin[2]

        # Order ascending position (DICOM: z increases from feet to head)
        return pd.DataFrame({
            "original_object_order": list(range(len(self.image_file_objects))),
            "position_z": image_position_z,
            "position_y": image_position_y,
            "position_x": image_position_x
        }).sort_values(by=["position_z", "position_y", "position_x"], ignore_index=True)

    def set_object_metadata(self):
        """
        Updates the object metadata that is passed to native image and mask classes in to_object. In this case,
        object metadata is copied from the underlying slices.
        """
        self.image_file_objects[0].set_object_metadata()
        self.object_metadata = self.image_file_objects[0].object_metadata
        self.object_metadata.pop("file_name", None)

    def export_metadata(self) -> None | dict[str, Any]:
        metadata = super().export_metadata()
        additional_metadata = self.image_file_objects[0].export_metadata(only_self=True)
        metadata.update(additional_metadata)

        return metadata

    def check_associated_masks(self):

        if self.associated_masks is None:
            return

        for mask in self.associated_masks:
            self._check_associated_mask_image_data(mask=mask)

    def _check_associated_mask_image_data(self, mask: MaskFile):
        """
        Check

        """
        from mirp._data_import.dicom_file import MaskDicomFile

        # Skip if the mask is a DICOM file.
        if isinstance(mask, MaskDicomFile):
            return

        super()._check_associated_mask_image_data(mask=mask)


class MaskDicomFileStack(ImageDicomFileStack, MaskFileStack):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)



import itk
import os.path
import copy
import numpy as np

from mirp._data_import.generic_file import ImageFile, MaskFile


class ImageITKFile(ImageFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def is_stackable(self, stack_images: str) -> bool:
        """
        Is the image potentially stackable?
        :param stack_images: One of auto, yes or no. By default (auto), images are not stackable. Images might be
        stackable if an image object represents a single slice.
        :return: boolean value
        """
        if stack_images == "auto":
            return False
        elif stack_images == "yes":
            if self.image_dimension is None:
                raise ValueError(
                    "DEV: The image_dimension argument is expected to be set. Call load_metadata to set this attribute."
                )

            if len(self.image_dimension) < 3:
                return True
            elif self.image_dimension[0] == 1:
                return True
            else:
                return False

        elif stack_images == "no":
            return False
        else:
            raise ValueError(
                f"The stack_images argument is expected to be one of yes, auto, or no. Found: {stack_images}. "
                f"[{self.describe_self()}]"
            )

    def copy(self):
        # Remove metadata during copying to prevent ITK SwigPyObjects from being pickled (which produces errors).
        temporary_metadata = self.image_metadata
        self.remove_metadata()

        # Create a copy without copying any metadata.
        image_copy = copy.deepcopy(self)

        # Restore metadata.
        self.image_metadata = temporary_metadata
        image_copy.image_metadata = temporary_metadata

        return image_copy

    def _complete_image_origin(self, force=False, frame_id=None):
        if self.image_origin is None:
            self.load_metadata()

            # Get dimensions, as this determines what can be meaningfully read.
            n_dimensions = self.image_metadata.GetNumberOfDimensions()

            # Get origin.
            origin = [0.0] * n_dimensions
            for ii in range(n_dimensions):
                origin[n_dimensions - (ii + 1)] = self.image_metadata.GetOrigin(ii)

            # Set origin.
            self.image_origin = tuple(origin)

    def _complete_image_orientation(self, force=False, frame_id=None):
        if self.image_orientation is None:
            self.load_metadata()

            # Get dimensions, as this determines what can be meaningfully read.
            n_dimensions = self.image_metadata.GetNumberOfDimensions()

            # Get orientation.
            orientation = []
            for ii in range(n_dimensions):
                orientation.append(self.image_metadata.GetDirection(ii))

            self.image_orientation = np.reshape(np.ravel(np.array(orientation))[::-1], [n_dimensions, n_dimensions])

    def _complete_image_spacing(self, force=False, frame_id=None):
        if self.image_spacing is None:
            self.load_metadata()

            # Get dimensions, as this determines what can be meaningfully read.
            n_dimensions = self.image_metadata.GetNumberOfDimensions()

            # Get spacing.
            spacing = [0.0] * n_dimensions
            for ii in range(n_dimensions):
                spacing[n_dimensions - (ii + 1)] = self.image_metadata.GetSpacing(ii)

            # Set spacing.
            self.image_spacing = tuple(spacing)

    def _complete_image_dimensions(self, force=False):
        self.load_metadata()

        if self.image_dimension is None:
            # Get dimensions, as this determines what can be meaningfully read.
            n_dimensions = self.image_metadata.GetNumberOfDimensions()

            # Get size of dimensions
            dimensions = [0] * n_dimensions
            for ii in range(n_dimensions):
                dimensions[n_dimensions - (ii + 1)] = self.image_metadata.GetDimensions(ii)

            # Set size of dimensions
            self.image_dimension = tuple(dimensions)

    def create(self):
        return self

    def load_metadata(self, **kwargs):
        if self.image_metadata is not None:
            return

        if self.file_path is None or not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path}. "
                f"[{self.describe_self()}]"
            )

        # Generate reader.
        reader = itk.ImageIOFactory.CreateImageIO(self.file_path, itk.CommonEnums.IOFileMode_ReadMode)
        reader.SetFileName(self.file_path)
        reader.ReadImageInformation()

        self.image_metadata = reader

    def load_data(self, **kwargs):
        if self.image_data is not None:
            return

        if self.file_path is None or not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path}. "
                f"[{self.describe_self()}]"
            )

        # Load the image
        itk_img = itk.imread(os.path.join(self.file_path))
        self.image_data = itk.GetArrayFromImage(itk_img).astype(np.float32)


class MaskITKFile(ImageITKFile, MaskFile):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def load_data(self, **kwargs):
        if self.image_data is not None:
            return

        if self.file_path is None or not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The mask file could not be found at the expected location: {self.file_path}. "
                f"[{self.describe_self()}]"
            )

        # Load the image
        itk_img = itk.imread(os.path.join(self.file_path))
        self.image_data = itk.GetArrayFromImage(itk_img).astype(int)



import copy

import numpy as np
import pandas as pd

from mirp._features.texture_matrix import DirectionalMatrix


class MatrixCM(DirectionalMatrix):

    def __init__(self, distance: int, **kwargs):
        super().__init__(**kwargs)

        # Set distance
        self.distance = int(distance)

        # Placeholders for derivative values computed using set_values_from_matrix
        # Co-occurrence matrix, expressed as probabilities.
        self.pij: pd.DataFrame | None = None

        # Marginal sum over grey levels of originating pixels i
        self.pi: pd.DataFrame | None = None

        # Marginal sum over grey levels of target pixels j
        self.pj: pd.DataFrame | None = None

        # Diagonal probabilities
        self.pimj: pd.DataFrame | None = None

        # Cross-diagonal probabilities
        self.pipj: pd.DataFrame | None = None

        # Number of intensities bins (grey levels)
        self.n_g: int | None = None

        # Mean co-occurrence weighted intensity
        self.mu: float | None = None

        # Marginal co-occurrence weighted intensity mean
        self.mu_marg: float | None = None

    def compute(
            self,
            data: pd.DataFrame | None = None,
            image_dimension: tuple[int, int, int] | None = None,
            **kwargs
    ):
        # Check if data actually exists
        if data is None:
            return

        # Check if the roi contains any masked voxels. If this is not the case, don't construct the GLRLM.
        if not np.any(data.roi_int_mask):
            return

        if image_dimension is None:
            raise ValueError("image_dimension cannot be None, but may not have been provided in the calling function.")

        # Create local copies of the image table
        if self.spatial_method in ["3d_average", "3d_volume_merge"]:
            data = copy.deepcopy(data)

        elif self.spatial_method in ["2d_average", "2d_slice_merge", "2.5d_direction_merge", "2.5d_volume_merge"]:
            data = copy.deepcopy(data[data.z == self.slice_id])
            data["index_id"] = np.arange(0, len(data))
            data["z"] = 0
            data = data.reset_index(drop=True)

        else:
            self._spatial_method_error()

        # Set grey level of voxels outside ROI to NaN
        data.loc[data.roi_int_mask == False, "g"] = np.nan

        # Determine potential transitions. The distance parameter determines the lookup distance.
        data["to_index"] = self.coord_to_index(
            x=data.x.values + self.direction[2] * self.distance,
            y=data.y.values + self.direction[1] * self.distance,
            z=data.z.values + self.direction[0] * self.distance,
            dims=image_dimension
        )

        # Look up intensity of neighbours.
        data["to_g"] = self._lookup_intensity(
            x=data.g.values,
            index=data.to_index.values
        )

        # Check if any valid neighbours were found.
        if np.all(np.isnan(data[["to_g"]])):
            return

        # Count occurrences intensity combinations.
        data = data.groupby(by=["g", "to_g"]).size().reset_index(name="n")

        # We assume symmetric GLCM. Hence, append grey level transitions in opposite direction.
        additional_data = pd.DataFrame({
            "g": data.to_g,
            "to_g": data.g,
            "n": data.n
        })
        data = pd.concat([data, additional_data], ignore_index=True)

        # Sum occurrences of grey level transitions after making the matrix symmetric.
        matrix = data.groupby(by=["g", "to_g"]).sum().reset_index()

        # Rename columns
        matrix.columns = ["i", "j", "n"]

        # Set the number of voxels
        self.n_voxels = np.sum(matrix.n)

        # Add matrix to object
        self.matrix = matrix

    def set_values_from_matrix(self, intensity_range: tuple[int, int], **kwargs):
        if self.is_empty():
            return

        self.pij = copy.deepcopy(self.matrix)
        self.pij["pij"] = self.pij.n / sum(self.pij.n)
        self.pi = self.pij.groupby(by="i")["pij"].sum().reset_index().rename(columns={"pij": "pi"})
        self.pj = self.pij.groupby(by="j")["pij"].sum().reset_index().rename(columns={"pij": "pj"})

        # Diagonal probabilities p(i-j)
        self.pimj = copy.deepcopy(self.pij)
        self.pimj["k"] = np.abs(self.pimj.i - self.pimj.j)
        self.pimj = self.pimj.groupby(by="k")["pij"].sum().reset_index().rename(columns={"pij": "pimj"})

        # Cross-diagonal probabilities p(i+j)
        self.pipj = copy.deepcopy(self.pij)
        self.pipj["k"] = self.pipj.i + self.pipj.j
        self.pipj = self.pipj.groupby(by="k")["pij"].sum().reset_index().rename(columns={"pij": "pipj"})

        # Merger of df.p_ij, df.p_i and df.p_j
        self.pij = pd.merge(self.pij, self.pi, on="i")
        self.pij = pd.merge(self.pij, self.pj, on="j")

        # Number of grey levels
        self.n_g = intensity_range[1] - intensity_range[0] + 1

        # Mean co-occurrence weighted intensity
        self.mu = np.sum(self.pij.i * self.pij.pij)

        # Marginal co-occurrence weighted intensity mean
        self.mu_marg = np.sum(self.pi.i * self.pi.pi)

    @staticmethod
    def _get_grouping_columns():
        return ["i", "j"]



from functools import lru_cache
from typing import Generator

import numpy as np

from mirp._features.ngldm_matrix import MatrixNGLDM
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.histogram import get_discretisation_parameters
from mirp._features.texture_features import FeatureTexture
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class FeatureNGLDM(FeatureTexture):

    def __init__(
            self,
            distance: float,
            coarseness: float,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Lookup distance (in voxels)
        self.distance = float(distance)

        # Coarseness
        self.coarseness = coarseness

        # Features are always computed from symmetric co-occurrence matrices.
        self.symmetric_matrix = True

        # Perform close crop for CM.
        self.cropping_distance = 0.0

    def get_matrix(
            self,
            image: GenericImage,
            mask: BaseMask,
    ) -> list[MatrixNGLDM]:
        # First discretise the image.
        image, mask = self.discretise_image(
            image=image,
            mask=mask,
            discretisation_method=self.discretisation_method,
            bin_width=self.bin_width,
            bin_number=self.bin_number,
            cropping_distance=self.cropping_distance
        )

        # Then get matrix or matrices
        matrix_list = self._get_matrix(
            image=image,
            mask=mask,
            spatial_method=self.spatial_method,
            distance=self.distance,
            coarseness=self.coarseness
        )

        return matrix_list

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_matrix(
            image: GenericImage,
            mask: BaseMask,
            spatial_method: str,
            distance: float,
            coarseness: float
    ) -> list[MatrixNGLDM]:
        # Represent image and mask as a dataframe.
        data = mask.as_pandas_dataframe(
            image=image,
            intensity_mask=True
        )

        # Instantiate a helper copy of the current class to be able to use class methods without tying the cache to the
        # instance of the original object from which this method is called.
        matrix_instance = MatrixNGLDM(
            spatial_method=spatial_method,
            distance=distance,
            coarseness=coarseness
        )

        # Compute the required matrices.
        matrix_list = list(matrix_instance.generate(
            prototype=MatrixNGLDM,
            n_slices=image.image_dimension[0],
            distance=distance,
            coarseness=coarseness
        ))
        for matrix in matrix_list:
            matrix.compute(data=data, image_dimension=image.image_dimension)

        # Merge according to the spatial method.
        matrix_list = matrix_instance.merge(
            matrix_list,
            prototype=MatrixNGLDM,
            distance=distance,
            coarseness=coarseness
        )

        # Compute additional values from the individual matrices.
        for matrix in matrix_list:
            matrix.set_values_from_matrix()

        return matrix_list

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureNGLDM):
            self._get_matrix.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_matrix.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Compute or retrieve matrices from cache.
        matrices = self.get_matrix(image=image, mask=mask)

        # Compute feature value from matrices, and average over matrices.
        values = [self._compute(matrix=matrix) for matrix in matrices]
        if np.all(np.isnan(values)):
            self.value = np.nan
        else:
            self.value = np.nanmean(values)

    @staticmethod
    def _compute(matrix: MatrixNGLDM):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = (
                self._get_base_table_name_element()
                + ["d" + str(int(np.round(self.distance, decimals=0)))]
                + ["a" + str(np.round(self.coarseness, decimals=0))]
                + self._get_spatial_table_name_element()
                + self._get_discretisation_table_name_element()
        )
        self.table_name = "_".join(table_elements)


class FeatureNGLDMLowDependenceEmphasis(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - low dependence emphasis"
        self.abbr_name = "ngl_lde"
        self.ibsi_id = "SODN"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sj.sj / matrix.sj.j ** 2.0) / matrix.n_s


class FeatureNGLDMHighDependenceEmphasis(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - high dependence emphasis"
        self.abbr_name = "ngl_hde"
        self.ibsi_id = "IMOQ"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sj.sj * matrix.sj.j ** 2.0) / matrix.n_s


class FeatureNGLDMLowGreyLevelCountEmphasis(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - low grey level count emphasis"
        self.abbr_name = "ngl_lgce"
        self.ibsi_id = "TL9H"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.si.si / matrix.si.i ** 2.0) / matrix.n_s


class FeatureNGLDMHighGreyLevelCountEmphasis(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - high grey level count emphasis"
        self.abbr_name = "ngl_hgce"
        self.ibsi_id = "OAE7"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.si.si * matrix.si.i ** 2.0) / matrix.n_s


class FeatureNGLDMLowDependenceLowGreyLevelEmphasis(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - ow dependence low grey level emphasis"
        self.abbr_name = "ngl_ldlge"
        self.ibsi_id = "EQ3F"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij / (matrix.sij.i * matrix.sij.j) ** 2.0) / matrix.n_s


class FeatureNGLDMLowDependenceHighGreyLevelEmphasis(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - low dependence high grey level emphasis"
        self.abbr_name = "ngl_ldhge"
        self.ibsi_id = "JA6D"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij * matrix.sij.i ** 2.0 / matrix.sij.j ** 2.0) / matrix.n_s


class FeatureNGLDMHighDependenceLowGreyLevelEmphasis(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - high dependence low grey level emphasis"
        self.abbr_name = "ngl_hdlge"
        self.ibsi_id = "NBZI"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij * matrix.sij.j ** 2.0 / matrix.sij.i ** 2.0) / matrix.n_s


class FeatureNGLDMHighDependenceHighGreyLevelEmphasis(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - high dependence high grey level emphasis"
        self.abbr_name = "ngl_hdhge"
        self.ibsi_id = "9QMG"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij * matrix.sij.i ** 2.0 * matrix.sij.j ** 2.0) / matrix.n_s


class FeatureNGLDMGreyLevelNonUniformity(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - grey level non-uniformity"
        self.abbr_name = "ngl_glnu"
        self.ibsi_id = "FP8K"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.si.si ** 2.0) / matrix.n_s


class FeatureNGLDMNormalisedGreyLevelNonUniformity(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - normalised grey level non-uniformity"
        self.abbr_name = "ngl_glnu_norm"
        self.ibsi_id = "5SPA"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.si.si ** 2.0) / matrix.n_s ** 2.0


class FeatureNGLDMDependenceCountNonUniformity(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - dependence count non-uniformity"
        self.abbr_name = "ngl_dcnu"
        self.ibsi_id = "Z87G"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sj.sj ** 2.0) / matrix.n_s


class FeatureNGLDMNormalisedDependenceCountNonUniformity(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - normalised dependence count non-uniformity"
        self.abbr_name = "ngl_dcnu_norm"
        self.ibsi_id = "OKJI"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sj.sj ** 2.0) / matrix.n_s ** 2.0


class FeatureNGLDMDependenceCountPercentage(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - dependence count percentage"
        self.abbr_name = "ngl_dc_perc"
        self.ibsi_id = "6XV8"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return matrix.n_s / matrix.n_voxels


class FeatureNGLDMGreyLevelVariance(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - grey level variance"
        self.abbr_name = "ngl_gl_var"
        self.ibsi_id = "1PFV"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.sij.sij * matrix.sij.i) / matrix.n_s
        return np.sum((matrix.sij.i - mu) ** 2.0 * matrix.sij.sij) / matrix.n_s


class FeatureNGLDMDependenceCountVariance(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - dependence count variance"
        self.abbr_name = "ngl_dc_var"
        self.ibsi_id = "DNX2"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        mu = np.sum(matrix.sij.sij * matrix.sij.j) / matrix.n_s
        return np.sum((matrix.sij.j - mu) ** 2.0 * matrix.sij.sij) / matrix.n_s


class FeatureNGLDMDependenceCountEntropy(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - dependence count entropy"
        self.abbr_name = "ngl_dc_entr"
        self.ibsi_id = "FCBV"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return -np.sum(matrix.sij.sij * np.log2(matrix.sij.sij / matrix.n_s)) / matrix.n_s


class FeatureNGLDMDependenceCountEnergy(FeatureNGLDM):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "NGLDM - dependence count energy"
        self.abbr_name = "ngl_dc_energy"
        self.ibsi_id = "CAS9"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(matrix: MatrixNGLDM) -> float:
        if matrix.is_empty():
            return np.nan
        return np.sum(matrix.sij.sij ** 2.0) / (matrix.n_s ** 2.0)


def get_ngldm_class_dict() -> dict[str, FeatureNGLDM]:
    class_dict = {
        "ngl_lde": FeatureNGLDMLowDependenceEmphasis,
        "ngl_hde": FeatureNGLDMHighDependenceEmphasis,
        "ngl_lgce": FeatureNGLDMLowGreyLevelCountEmphasis,
        "ngl_hgce": FeatureNGLDMHighGreyLevelCountEmphasis,
        "ngl_ldlge": FeatureNGLDMLowDependenceLowGreyLevelEmphasis,
        "ngl_ldhge": FeatureNGLDMLowDependenceHighGreyLevelEmphasis,
        "ngl_hdlge": FeatureNGLDMHighDependenceLowGreyLevelEmphasis,
        "ngl_hdhge": FeatureNGLDMHighDependenceHighGreyLevelEmphasis,
        "ngl_glnu": FeatureNGLDMGreyLevelNonUniformity,
        "ngl_glnu_norm": FeatureNGLDMNormalisedGreyLevelNonUniformity,
        "ngl_dcnu": FeatureNGLDMDependenceCountNonUniformity,
        "ngl_dcnu_norm": FeatureNGLDMNormalisedDependenceCountNonUniformity,
        "ngl_dc_perc": FeatureNGLDMDependenceCountPercentage,
        "ngl_gl_var": FeatureNGLDMGreyLevelVariance,
        "ngl_dc_var": FeatureNGLDMDependenceCountVariance,
        "ngl_dc_entr": FeatureNGLDMDependenceCountEntropy,
        "ngl_dc_energy": FeatureNGLDMDependenceCountEnergy
    }

    return class_dict


def generate_ngldm_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureNGLDM, None, None]:
    class_dict = get_ngldm_class_dict()
    ngldm_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_ngldm_family():
        features = ngldm_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only NGLDM-features, and return if none are present.
    features = [feature for feature in features if feature in ngldm_features]
    if len(features) == 0:
        return

    # Features are parametrised by the choice of discretisation parameters, spatial methods, distance and coarseness.
    for discretisation_parameters in get_discretisation_parameters(
        settings=settings
    ):
        for spatial_method in settings.ngldm_spatial_method:
            for distance in settings.ngldm_dist:
                for coarseness in settings.ngldm_diff_lvl:
                    for feature in features:
                        yield class_dict[feature](
                            spatial_method=spatial_method,
                            distance=distance,
                            coarseness=coarseness,
                            **discretisation_parameters
                        )



import numpy as np
import copy

from mirp._images.generic_image import GenericImage
from mirp._images.transformed_image import GaborTransformedImage
from mirp._imagefilters.generic import GenericFilter
from mirp._imagefilters.utilities import pool_voxel_grids, FilterSet2D
from mirp.settings.generic import SettingsClass


class GaborFilter(GenericFilter):

    def __init__(self, image: GenericImage, settings: SettingsClass, name: str):

        super().__init__(image=image, settings=settings, name=name)
        self.ibsi_compliant = True
        self.ibsi_id = "Q88H"

        # Sigma parameter that determines filter width.
        self.sigma: None | float | list[float] = settings.img_transform.gabor_sigma

        # Eccentricity parameter
        self.gamma: None | float | list[float] = settings.img_transform.gabor_gamma

        # Wavelength parameter
        self.lambda_parameter: None | float | list[float] = settings.img_transform.gabor_lambda

        # Initial angle.
        self.theta: None | float | list[float] | int | list[int] = settings.img_transform.gabor_theta

        # Set whether theta is considered separate, or pooled.
        self.pool_theta: bool = settings.img_transform.gabor_pool_theta

        # Update ype of response
        self.response_type = settings.img_transform.gabor_response

        # Rotational invariance.
        self.rotation_invariance = settings.img_transform.gabor_rotation_invariance

        # Which pooling method is used.
        self.pooling_method = settings.img_transform.gabor_pooling_method

        # Boundary conditions.
        self.mode = settings.img_transform.gabor_boundary_condition

        # Riesz transformation settings.
        self.riesz_order: None | list[int] | list[list[int]] = None
        self.riesz_steered: bool = False
        self.riesz_sigma: None | float | list[float] = None
        if settings.img_transform.has_riesz_filter(x=name):
            self.riesz_order = settings.img_transform.riesz_order

            if settings.img_transform.has_steered_riesz_filter(x=name):
                self.riesz_steered = True
                self.riesz_sigma = settings.img_transform.riesz_filter_tensor_sigma

            # Riesz transformed filters are not IBSI-compliant
            self.ibsi_compliant = False

        # Set the axis orthogonal to the plane in which the Gabor kernel is applied.
        if self.separate_slices or not self.rotation_invariance:
            self.stack_axis: int | list[int] = [0]
        else:
            self.stack_axis: int | list[int] = [0, 1, 2]

    def generate_object(self, allow_pooling: bool = True):
        # Generator for transformation objects.
        sigma = copy.deepcopy(self.sigma)
        if not isinstance(sigma, list):
            sigma = [sigma]

        gamma = copy.deepcopy(self.gamma)
        if not isinstance(gamma, list):
            gamma = [gamma]

        lambda_p = copy.deepcopy(self.lambda_parameter)
        if not isinstance(lambda_p, list):
            lambda_p = [lambda_p]

        theta = copy.deepcopy(self.theta)
        if not isinstance(theta, list):
            theta = [theta]

        # Nest theta for internal iterations.
        if self.pool_theta and allow_pooling:
            theta = [theta]

        axis = copy.deepcopy(self.stack_axis)
        if not isinstance(axis, list):
            axis = [axis]

        # Nest axis for internal iterations.
        if self.pool_theta and allow_pooling:
            axis = [axis]

        riesz_order = copy.deepcopy(self.riesz_order)
        if riesz_order is None:
            riesz_order = [None]
        elif not all(isinstance(riesz_order_set, list) for riesz_order_set in riesz_order):
            riesz_order = [riesz_order]

        riesz_sigma = copy.deepcopy(self.riesz_sigma)
        if not isinstance(riesz_sigma, list):
            riesz_sigma = [riesz_sigma]

        # Iterate over options to yield filter objects with specific settings. A copy of the parent object is made to
        # avoid updating by reference.
        for current_sigma in sigma:
            for current_gamma in gamma:
                for current_lambda in lambda_p:
                    for current_riesz_order in riesz_order:
                        for current_riesz_sigma in riesz_sigma:
                            for current_theta in theta:
                                for current_axis in axis:

                                    filter_object = copy.deepcopy(self)
                                    filter_object.sigma = current_sigma
                                    filter_object.gamma = current_gamma
                                    filter_object.lambda_parameter = current_lambda
                                    filter_object.riesz_order = current_riesz_order
                                    filter_object.riesz_sigma = current_riesz_sigma
                                    filter_object.theta = current_theta
                                    filter_object.stack_axis = current_axis

                                    yield filter_object

    def transform(self, image: GenericImage) -> GaborTransformedImage:
        # Create placeholder Gabor response map.
        response_map = GaborTransformedImage(
            image_data=None,
            sigma_parameter=self.sigma,
            gamma_parameter=self.gamma,
            lambda_parameter=self.lambda_parameter,
            theta_parameter=self.theta,
            pool_theta=self.pool_theta,
            response_type=self.response_type,
            rotation_invariance=self.rotation_invariance,
            pooling_method=self.pooling_method,
            boundary_condition=self.mode,
            riesz_order=self.riesz_order,
            riesz_steering=self.riesz_steered,
            riesz_sigma_parameter=self.riesz_sigma,
            template=image
        )
        response_map.ibsi_compliant = self.ibsi_compliant and image.ibsi_compliant

        if image.is_empty():
            return response_map

        # Set response voxel grid.
        response_voxel_grid = None

        # Initialise iterator ii to avoid IDE warnings.
        ii = 0
        for ii, pooled_filter_object in enumerate(self.generate_object(allow_pooling=False)):
            # Generate transformed voxel grid.
            pooled_voxel_grid = pooled_filter_object.transform_grid(
                voxel_grid=image.get_voxel_grid(),
                spacing=np.array(image.image_spacing))

            # Pool voxel grids.
            response_voxel_grid = pool_voxel_grids(
                x1=response_voxel_grid,
                x2=pooled_voxel_grid,
                pooling_method=self.pooling_method)

        if self.pooling_method == "mean":
            # Perform final pooling step for mean pooling.
            response_voxel_grid = np.divide(response_voxel_grid, ii + 1)

        # Set voxel grid.
        response_map.set_voxel_grid(voxel_grid=response_voxel_grid)

        return response_map

    def transform_grid(
            self,
            voxel_grid: np.ndarray,
            spacing: np.array):

        # Get in-plane spacing, i.e. not stack_axis.
        spacing: float = max([
            current_spacing for ii, current_spacing in enumerate(spacing.tolist())
            if not ii == self.stack_axis
        ])

        # Convert sigma from physical units to voxel units.
        sigma: float = self.sigma / spacing
        lambda_p: float = self.lambda_parameter / spacing

        # Convert theta to radians.
        theta = np.deg2rad(self.theta)

        # Get size of the voxelgrid as filter size.
        x_size = y_size = max([
            current_shape for ii, current_shape in enumerate(voxel_grid.shape)
            if not ii == self.stack_axis
        ])

        # Ensure that size is uneven.
        x_size = int(1 + 2 * np.floor(x_size / 2.0))
        y_size = int(1 + 2 * np.floor(y_size / 2.0))

        # Create grid coordinates with [0, 0] in the center.
        y, x = np.mgrid[:y_size, :x_size].astype(float)
        y -= (y_size - 1.0) / 2.0
        x -= (x_size - 1.0) / 2.0

        # Compute rotation matrix: Since we are computing clock-wise rotations, use negative angles.
        rotation_matrix = np.array([[-np.cos(theta), np.sin(theta)],
                                    [np.sin(theta), np.cos(theta)]])

        # Compute rotated grid coordinates around the center.
        rotated_scan_coordinates = np.dot(rotation_matrix, np.array((y.flatten(), x.flatten())))
        y = rotated_scan_coordinates[0, :].reshape((y_size, x_size))
        x = rotated_scan_coordinates[1, :].reshape((y_size, x_size))

        # Create filter weights.
        gabor_filter = np.exp(-(np.power(x, 2.0) + self.gamma ** 2.0 * np.power(y, 2.0)) / (2.0 * sigma ** 2.0) + 1.0j
                              * (2.0 * np.pi * x) / lambda_p)

        # Create filter
        gabor_filter = FilterSet2D(
            gabor_filter,
            riesz_order=self.riesz_order,
            riesz_steered=self.riesz_steered,
            riesz_sigma=self.riesz_sigma)

        # Convolve gabor filter with the image.
        response_map = gabor_filter.convolve(
            voxel_grid=voxel_grid,
            mode=self.mode,
            response=self.response_type,
            axis=self.stack_axis)

        # Compute the convolution
        return response_map



import warnings
from functools import lru_cache
from typing import Generator

import numpy as np
import pandas as pd
import scipy.ndimage as ndi


from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp._features.base_feature import Feature
from mirp._features.utilities import rep
from mirp.settings.feature_parameters import FeatureExtractionSettingsClass


class DataLocalIntensity(object):

    def __init__(self):
        # Raw intensity data
        self.data: pd.DataFrame | None = None

        # Number of voxels
        self.n_voxels: int | None = None

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_intensity.is_empty_mask():
            return

        # Set number of voxels.
        self.n_voxels = np.sum(mask.roi_intensity.get_voxel_grid())

        # Set data.
        if self.n_voxels > 300:
            self._filter_compute(image=image, mask=mask)
        else:
            self._direct_compute(image=image, mask=mask)

        if not self.is_empty():
            # Shrink to contain only voxels within the intensity mask.
            self.data = self.data.loc[self.data.in_roi == True, :]

    def _filter_compute(self, image: GenericImage, mask: BaseMask):
        # Determine distance
        distance = (3.0 / (4.0 * np.pi)) ** (1.0 / 3.0) * 10.0

        # Get maximal extension in cubic space
        base_ext = np.floor(distance / np.array(image.image_spacing))

        # Create displacement map
        df_base = pd.DataFrame({
            "x": rep(
                x=np.arange(-base_ext[2], base_ext[2] + 1),
                each=(2 * base_ext[0] + 1) * (2 * base_ext[1] + 1),
                times=1),
            "y": rep(
                x=np.arange(-base_ext[1], base_ext[1] + 1),
                each=2 * base_ext[0] + 1,
                times=2 * base_ext[2] + 1),
            "z": rep(
                x=np.arange(-base_ext[0], base_ext[0] + 1),
                each=1,
                times=(2 * base_ext[1] + 1) * (2 * base_ext[2] + 1))
        })

        # Calculate distances for displacement map
        df_base["dist"] = np.sqrt(np.sum(np.multiply(
            df_base.loc[:, ("z", "y", "x")].values, image.image_spacing
        ) ** 2.0, axis=1))

        # Identify elements in range
        df_base["set_weight"] = df_base.dist <= distance

        # Set weights for filter
        df_base["weight"] = np.zeros(len(df_base))
        df_base.loc[df_base.set_weight == True, "weight"] = 1.0 / np.sum(df_base.set_weight)

        # Update coordinates to start at 0
        df_base["x"] += base_ext[2]
        df_base["y"] += base_ext[1]
        df_base["z"] += base_ext[0]

        # Generate convolution filter
        conv_filter = np.zeros(shape=(
            int(np.max(df_base.z)) + 1,
            int(np.max(df_base.y)) + 1,
            int(np.max(df_base.x)) + 1)
        )
        conv_filter[df_base.z.astype(int), df_base.y.astype(int), df_base.x.astype(int)] = df_base.weight

        try:
            # Filter image using mean filter
            if image.get_default_lowest_intensity() is not None:
                # Use 0.0 constant for PET data
                img_avg = ndi.convolve(
                    image.get_voxel_grid(),
                    weights=conv_filter,
                    mode="constant",
                    cval=image.get_default_lowest_intensity()
                )
            else:
                img_avg = ndi.convolve(
                    image.get_voxel_grid(),
                    weights=conv_filter,
                    mode="nearest"
                )
        except MemoryError as err:
            warnings.warn(
                f"Unable to compute local intensity features due to Memory error. The image size and "
                f"resolution may be too small to fit a 1cm³ spherical filter for computing local intensity features.",
                UserWarning
            )
            return

        # Construct data frame for comparison
        self.data = pd.DataFrame({
            "g": np.ravel(image.get_voxel_grid()),
            "g_loc": np.ravel(img_avg),
            "in_roi": np.ravel(mask.roi_intensity.get_voxel_grid())
        })

    def _direct_compute(self, image: GenericImage, mask: BaseMask):
        # Determine distance
        distance = (3.0 / (4.0 * np.pi)) ** (1.0 / 3.0) * 10.0

        # Construct data frame for comparison
        self.data = pd.DataFrame({
            "g": np.ravel(image.get_voxel_grid()),
            "g_loc": np.ravel(np.full(image.image_dimension, np.nan)),
            "in_roi": np.ravel(mask.roi_intensity.get_voxel_grid())
        })

        # Generate position matrix
        pos_mat = np.array(
            np.unravel_index(indices=np.arange(0, np.prod(image.image_dimension)), shape=image.image_dimension),
            dtype=np.float32
        ).transpose()

        # Iterate over voxels in the roi
        if np.sum(self.data.in_roi) > 1:
            for i in np.array(np.where(self.data.in_roi == True)).squeeze():
                # Determine distance from currently selected voxel
                vox_dist = np.sqrt(
                    np.sum(np.power(np.multiply(pos_mat - pos_mat[i, :], image.image_spacing), 2.0), axis=1))

                # Calculate mean grey level over all voxels within range
                self.data.loc[i, "g_loc"] = np.mean(self.data.g[vox_dist <= distance])

        else:
            i = np.where(self.data.in_roi == True)[0][0]

            # Determine distance from currently selected voxel
            vox_dist = np.sqrt(np.sum(np.power(np.multiply(pos_mat - pos_mat[i, :], image.image_spacing), 2.0), axis=1))

            # Calculate mean grey level over all voxels within range
            self.data.loc[i, "g_loc"] = np.mean(self.data.g[vox_dist <= distance])

    def is_empty(self):
        return self.data is None


class FeatureLocalIntensity(Feature):

    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)

    @staticmethod
    @lru_cache(maxsize=1)
    def _get_data(
            image: GenericImage,
            mask: BaseMask
    ) -> DataLocalIntensity:
        data = DataLocalIntensity()
        data.compute(image=image, mask=mask)

        return data

    def clear_local_cache(self, other):
        if not isinstance(other, FeatureLocalIntensity):
            self._get_data.cache_clear()

    def clear_cache(self):
        super().clear_cache()
        self._get_data.cache_clear()

    def compute(self, image: GenericImage, mask: BaseMask):
        # Get data.
        data = self._get_data(image=image, mask=mask)

        # Compute feature value.
        if data.is_empty():
            self.value = np.nan
        else:
            self.value = self._compute(data=data)

    @staticmethod
    def _compute(data: DataLocalIntensity):
        raise NotImplementedError("Implement _compute for feature-specific computation.")

    def create_table_name(self):
        table_elements = self._get_base_table_name_element()
        self.table_name = "_".join(table_elements)


class FeatureLocalIntensityLocalPeak(FeatureLocalIntensity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Local intensity - local peak"
        self.abbr_name = "loc_peak_loc"
        self.ibsi_id = "VJGA"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataLocalIntensity) -> float:
        return np.max(data.data.loc[data.data.g == np.max(data.data.g), "g_loc"])


class FeatureLocalIntensityGlobalPeak(FeatureLocalIntensity):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = "Local intensity - global peak"
        self.abbr_name = "loc_peak_glob"
        self.ibsi_id = "0F91"
        self.ibsi_compliant = True

    @staticmethod
    def _compute(data: DataLocalIntensity) -> float:
        return np.max(data.data.g_loc)


def get_local_intensity_class_dict() -> dict[str, FeatureLocalIntensity]:
    class_dict = {
        "loc_peak_loc": FeatureLocalIntensityLocalPeak,
        "loc_peak_glob": FeatureLocalIntensityGlobalPeak
    }

    return class_dict


def generate_local_intensity_features(
        settings: FeatureExtractionSettingsClass,
        features: None | list[str] = None
) -> Generator[FeatureLocalIntensity, None, None]:
    class_dict = get_local_intensity_class_dict()
    local_int_features = list(class_dict.keys())

    # Populate features if available.
    if features is None and settings.has_local_intensity_family():
        features = local_int_features

    # Terminate early if no features are set, and none are required.
    if features is None:
        return

    # Select only local intensity features, and return if none are present.
    features = [feature for feature in features if feature in local_int_features]
    if len(features) == 0:
        return

    for feature in features:
        yield class_dict[feature]()


import copy

import numpy as np
import pandas as pd

from mirp._features.texture_matrix import DirectionalMatrix


class MatrixRLM(DirectionalMatrix):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Placeholders for derivative values computed using set_values_from_matrix
        # Run-length matrix
        self.rij: pd.DataFrame | None = None

        # Marginal sum over grey levels
        self.ri: pd.DataFrame | None = None

        # Marginal sum over run lengths
        self.rj: pd.DataFrame | None = None

        # Number of runs
        self.n_s: int | None = None

    def compute(
            self,
            data: pd.DataFrame | None = None,
            image_dimension: tuple[int, int, int] | None = None,
            **kwargs
    ):
        # Check if data actually exists
        if data is None:
            return

        # Check if the roi contains any masked voxels. If this is not the case, don't construct the GLRLM.
        if not np.any(data.roi_int_mask):
            return

        if image_dimension is None:
            raise ValueError("image_dimension cannot be None, but may not have been provided in the calling function.")

        # Create local copies of the image table
        if self.spatial_method in ["3d_average", "3d_volume_merge"]:
            data = copy.deepcopy(data)

        elif self.spatial_method in ["2d_average", "2d_slice_merge", "2.5d_direction_merge", "2.5d_volume_merge"]:
            data = copy.deepcopy(data[data.z == self.slice_id])
            data["index_id"] = np.arange(0, len(data))
            data["z"] = 0
            data = data.reset_index(drop=True)

        else:
            self._spatial_method_error()

        # Set grey level of voxels outside ROI to NaN
        data.loc[data.roi_int_mask == False, "g"] = np.nan

        # Set the number of voxels
        self.n_voxels = np.sum(data.roi_int_mask.values)

        # Determine update index number for direction
        if ((
                self.direction[2]
                + self.direction[1] * image_dimension[2]
                + self.direction[0] * image_dimension[2] * image_dimension[1]
        ) >= 0):
            direction = self.direction
        else:
            direction = tuple(-x for x in self.direction)

        # Step size
        index_update = (
                direction[2]
                + direction[1] * image_dimension[2]
                + direction[0] * image_dimension[2] * image_dimension[1]
        )

        # Generate information concerning segments
        n_segments = index_update  # Number of segments

        # Check if the number of segments is greater than one
        if n_segments == 0:
            return

        # Nominal segment length
        segment_length = (len(data) - 1) // index_update + 1

        # Initial segment length for transitions (nominal length - 1)
        trans_segment_length = np.tile([segment_length - 1], reps=n_segments)

        # Number of full segments
        full_len_trans = n_segments - n_segments * segment_length + len(data)

        # Update full segments
        trans_segment_length[0:full_len_trans] += 1

        # Create transition vector
        trans_vec = (
                np.tile(
                    np.arange(start=0, stop=len(data), step=index_update),
                    reps=index_update
                )
                + np.repeat(
                    np.arange(start=0, stop=n_segments),
                    repeats=segment_length
                )
            )
        trans_vec = trans_vec[trans_vec < len(data)]

        # Determine valid transitions
        to_index = self.coord_to_index(
            x=data.x.values + direction[2],
            y=data.y.values + direction[1],
            z=data.z.values + direction[0],
            dims=image_dimension
        )

        # Determine which transitions are valid
        end_ind = np.nonzero(to_index[trans_vec] < 0)[0]  # Find transitions that form an endpoint.

        # Get an interspersed array of intensities. Runs are broken up by np.nan
        intensities = np.insert(data.g.values[trans_vec], end_ind + 1, np.nan)

        # Determine run length start and end indices
        rle_end = np.array(np.append(np.where(intensities[1:] != intensities[:-1]), len(intensities) - 1))
        rle_start = np.cumsum(np.append(0, np.diff(np.append(-1, rle_end))))[:-1]

        # Generate matrix
        matrix = pd.DataFrame({
            "i": intensities[rle_start],
            "j": rle_end - rle_start + 1
        })
        matrix = matrix.loc[~np.isnan(matrix.i), :]
        matrix = matrix.groupby(by=["i", "j"]).size().reset_index(name="rij")

        # Add matrix to object
        self.matrix = matrix

    def set_values_from_matrix(self, **kwargs):
        if self.is_empty():
            return

        # Copy of matrix
        self.rij = copy.deepcopy(self.matrix)

        # Sum over grey levels
        self.ri = self.matrix.groupby(by="i")["rij"].sum().reset_index().rename(columns={"rij": "ri"})

        # Sum over run lengths
        self.rj = self.matrix.groupby(by="j")["rij"].sum().reset_index().rename(columns={"rij": "rj"})

        # Constant definitions
        self.n_s = np.sum(self.matrix.rij) * 1.0  # Number of runs

    @staticmethod
    def _get_grouping_columns():
        return ["i", "j"]



from mirp._data_import.generic_file import MaskFile
from mirp.data_import.import_image import _import_image
from mirp._data_import.utilities import supported_mask_modalities, supported_file_types, flatten_list


def import_mask(
        mask,
        sample_name: None | str | list[str] = None,
        mask_name: None | str | list[str] = None,
        mask_file_type: None | str = None,
        mask_modality: None | str | list[str] = None,
        mask_sub_folder: None | str = None,
        roi_name: None | str | list[str] | dict[str, str] = None,
        stack_masks: str = "auto"
) -> list[MaskFile]:
    """
    Creates and curates references to mask files. Masks determine the location of regions of interest. They are
    usually created by manual, semi-automatic or fully automatic segmentation.

    Parameters
    ----------
    mask: Any
        A path to a mask file, a path to a directory containing mask files, a path to a config_data.xml
        file, a path to a csv file containing references to mask files, a pandas.DataFrame containing references to
        mask files, or a numpy.ndarray.

    sample_name: str or list of str, optional, default: None
        Name of expected sample names. This is used to select specific mask files. If None, no mask files are filtered
         based on the corresponding sample name (if known).

    mask_name: str, optional, default: None
        Pattern to match mask files against. The matches are exact. Use wildcard symbols ("*") to match varying
        structures. The sample name (if part of the file name) can also be specified using "#". For example,
        mask_name = '#_*_mask' would find John_Doe in John_Doe_CT_mask.nii or John_Doe_001_mask.nii. File extensions
        do not need to be specified. If None, file names are not used for filtering files and setting sample names.

    mask_file_type: {"dicom", "nifti", "nrrd", "numpy", "itk"}, optional, default: None
        The type of file that is expected. If None, the file type is not used for filtering files.
        "itk" comprises "nifti" and "nrrd" file types.

    mask_modality: {"rtstruct", "seg", "generic_mask"}, optional, default: None
        The type of modality that is expected. If None, modality is not used for filtering files.
        Note that only DICOM files contain metadata concerning modality. Masks from non-DICOM files are considered to
        be "generic_mask".

    mask_sub_folder: str, optional, default: None
        Fixed directory substructure where mask files are located. If None, the directory substructure is not used for
        filtering files.

    roi_name: str or list of str or dict, optional, default: None
        Name of the regions of interest that should be assessed.

    stack_masks: {"auto", "yes", "no"}, optional, default: "str"
        If mask files in the same directory cannot be assigned to different samples, and are 2D (slices) of the same
        size, they might belong to the same 3D mask stack. "auto" will stack 2D numpy arrays, but not other file
        types. "yes" will stack all files that contain 2D images, that have the same dimensions, orientation and
        spacing, except for DICOM files. "no" will not stack any files. DICOM files ignore this argument,
        because their stacking can be determined from metadata.

    Returns
    -------
    list of MaskFile
        The functions returns a list of MaskFile objects, if any were found with the specified filters.

    """
    # Check modality.
    if mask_modality is not None:
        if not isinstance(mask_modality, str):
            raise TypeError(
                f"The mask_modality argument is expected to be a single character string or None. The following "
                f"modalities are supported: {', '.join(supported_mask_modalities(None))}.")
        _ = supported_mask_modalities(mask_modality.lower())

    # Check image_file_type.
    if mask_file_type is not None:
        if not isinstance(mask_file_type, str):
            raise TypeError(
                f"The mask_file_type argument is expected to be a single character string, or None. The following file "
                f"types are supported: {', '.join(supported_file_types(None))}.")
        _ = supported_file_types(mask_file_type)

    # Check stack_images
    if stack_masks not in ["yes", "auto", "no"]:
        raise ValueError(
            f"The stack_images argument is expected to be one of yes, auto, or no. Found: {stack_masks}."
        )

    mask_list = _import_image(
        mask,
        sample_name=sample_name,
        image_name=mask_name,
        image_file_type=mask_file_type,
        image_modality=mask_modality,
        image_sub_folder=mask_sub_folder,
        stack_images=stack_masks,
        roi_name=roi_name,
        is_mask=True
    )

    if not isinstance(mask_list, list):
        mask_list = [mask_list]

    # Flatten list.
    mask_list = flatten_list(mask_list)

    return mask_list



import os.path
import numpy as np

from mirp._data_import.dicom_file import ImageDicomFile
from mirp._data_import.utilities import get_pydicom_meta_tag


class ImageDicomMultiFrame(ImageDicomFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def create(self):
        # This method is called from ImageDicomFile.create amd dispatches to modality-specific multi-frame objects.
        from mirp._data_import.dicom_file_ct import ImageDicomFileCTMultiFrame
        from mirp._data_import.dicom_file_pet import ImageDicomFilePTMultiFrame
        from mirp._data_import.dicom_file_mr import ImageDicomFileMRMultiFrame
        from mirp._data_import.dicom_file_mr_adc import ImageDicomFileMRADCMultiFrame
        from mirp._data_import.dicom_file_mr_dce import ImageDicomFileMRDCEMultiFrame

        if self.modality == "ct":
            file_class = ImageDicomFileCTMultiFrame
        elif self.modality == "pt":
            file_class = ImageDicomFilePTMultiFrame
        elif self.modality == "mr":
            file_class = ImageDicomFileMRMultiFrame
        elif self.modality == "adc":
            file_class = ImageDicomFileMRADCMultiFrame
        elif self.modality == "dce":
            file_class = ImageDicomFileMRDCEMultiFrame

        else:
            # Multi-frame is not implemented for the following modalities:
            # - RT Dose: lack of DICOM module for RT Dose with multi-frame data.
            raise NotImplementedError(
                f"Multi-frame DICOM not implemented for {self.modality} modality. Contact the devs."
            )

        image = file_class()
        image.update_from_template(template=self)

        return image

    def is_stackable(self, stack_images: str):
        # Multi-frame images might be actually be stackable (concatenated), but ignore that for now.
        return False

    def _complete_image_origin(self, force=False, frame_id=None):
        if self.image_origin is None:

            # Load relevant metadata.
            self.load_metadata(limited=True)

            if frame_id is None:
                frame_id = 0

            origin = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0020, 0x0032),
                tag_type="mult_float",
                macro_dcm_seq=(0x0020, 0x9113),
                frame_id=frame_id
            )[::-1]
            self.image_origin = tuple(origin)

    def _complete_image_orientation(self, force=False, frame_id=None):
        if self.image_orientation is None:

            # Load relevant metadata.
            self.load_metadata(limited=True)

            if frame_id is None:
                frame_id = 0

            orientation: list[float] = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0020, 0x0037),
                tag_type="mult_float",
                macro_dcm_seq=(0x0020, 0x9116),
                frame_id=frame_id
            )

            # First compute z-orientation.
            # noinspection PyUnreachableCode
            orientation += list(np.cross(orientation[0:3], orientation[3:6]))
            self.image_orientation = np.reshape(orientation[::-1], [3, 3], order="F")

    def _complete_image_spacing(self, force=False, frame_id=None):
        if self.image_spacing is None:
            # Load relevant metadata.
            self.load_metadata(limited=True)

            if frame_id is None:
                frame_id = 0

            # Get pixel-spacing.
            spacing = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0028, 0x0030),
                tag_type="mult_float",
                macro_dcm_seq=(0x0028, 0x9110),
                frame_id=frame_id
            )

            # First try to get spacing between slices.
            z_spacing = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0018, 0x0088),
                tag_type="float",
                macro_dcm_seq=(0x0028, 0x9110),
                frame_id=frame_id
            )

            # If spacing between slices is not set, get slice thickness.
            if z_spacing is None:
                z_spacing = get_pydicom_meta_tag(
                    dcm_seq=self.image_metadata,
                    tag=(0x0018, 0x0050),
                    tag_type="float",
                    macro_dcm_seq=(0x0028, 0x9110),
                    frame_id=frame_id
                )

            # If slice thickness is not set, use a default value.
            if z_spacing is None:
                z_spacing = 1.0
            spacing += [z_spacing]

            self.image_spacing = tuple(spacing[::-1])

    def _complete_image_dimensions(self, force=False):
        if self.image_dimension is None:
            # Load relevant metadata.
            self.load_metadata(limited=True)

            dimensions = tuple([
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0008), tag_type="int"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0010), tag_type="int"),
                get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0011), tag_type="int")
            ])

            self.image_dimension = dimensions

    def load_data(self, **kwargs):
        if self.image_data is not None:
            return self.image_data

        if self.file_path is not None and not os.path.exists(self.file_path):
            raise FileNotFoundError(
                f"The image file could not be found at the expected location: {self.file_path}. [{self.describe_self()}]"
            )

        if self.file_path is None:
            raise ValueError(f"A path to a file was expected, but not present. [{self.describe_self()}]")

        # Load metadata.
        self.load_metadata(include_image=True)
        image_data = self.image_metadata.pixel_array.astype(np.float32)

        # Determine number of frames
        n_frames = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0028, 0x0008), tag_type="int")

        # Determine rescale intercept.
        rescale_intercept = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0028, 0x1052),
            tag_type="float",
            macro_dcm_seq=(0x0028, 0x9145)
        )
        if rescale_intercept is None:
            if n_frames is None:
                rescale_intercept = 0.0
            else:
                rescale_intercept = [
                    get_pydicom_meta_tag(
                        dcm_seq=self.image_metadata,
                        tag=(0x0028, 0x1052),
                        tag_type="float",
                        macro_dcm_seq=(0x0028, 0x9145),
                        frame_id=frame_id,
                        default=0.0
                    )
                    for frame_id in np.arange(self.image_dimension[0])
                ]

        # Determine rescale slope.
        rescale_slope = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0028, 0x1053),
            tag_type="float",
            macro_dcm_seq=(0x0028, 0x9145)
        )
        if rescale_slope is None:
            if n_frames is None:
                rescale_slope = 1.0
            else:
                rescale_slope = [
                    get_pydicom_meta_tag(
                        dcm_seq=self.image_metadata,
                        tag=(0x0028, 0x1052),
                        tag_type="float",
                        macro_dcm_seq=(0x0028, 0x9145),
                        frame_id=frame_id,
                        default=1.0
                    )
                    for frame_id in np.arange(self.image_dimension[0])
                ]

        # Apply slope and intercept.
        if isinstance(rescale_slope, list):
            for ii, b in enumerate(rescale_slope):
                image_data[ii, :, :] = b * image_data[ii, :, :]
        else:
            image_data *= rescale_slope

        if isinstance(rescale_intercept, list):
            for ii, a in enumerate(rescale_intercept):
                image_data[ii, :, :] = image_data[ii, :, :] + a
        else:
            image_data += rescale_intercept

        self.image_data = image_data



import warnings

import numpy as np
import pandas as pd
import copy
import sys
from typing import Any
from pathlib import Path

from mirp._images.generic_image import GenericImage
from mirp._images.mask_image import MaskImage
from mirp.settings.generic import SettingsClass

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class BaseMask:
    def __init__(
            self,
            roi_name: str,
            **kwargs
    ):
        # Make cooperative.
        super().__init__()

        # Set region of interest.
        self.roi = MaskImage(**kwargs)

        # Set sample name.
        self.sample_name = self.roi.sample_name

        # Define other types of masks.
        self.roi_intensity: None | MaskImage = None
        self.roi_morphology: None | MaskImage = None

        # Set name of the mask.
        self.roi_name: str | list[str] = roi_name

        # Set intensity range.
        self.intensity_range: tuple[Any, Any] = tuple([np.nan, np.nan])

    @staticmethod
    def get_dir_path():
        # BaseMask does not have an associated directory path.
        return None

    @staticmethod
    def get_file_name():
        # BaseMask also has no associated file name.
        return None

    def get_image_origin(self, **kwargs):
        return self.roi.get_image_origin(**kwargs)

    def get_image_orientation(self, **kwargs):
        return self.roi.get_image_orientation(**kwargs)

    def get_image_spacing(self, **kwargs):
        return self.roi.get_image_spacing(**kwargs)

    def get_image_dimension(self, **kwargs):
        return self.roi.get_image_dimension(**kwargs)

    def remove_metadata(self, force=False):
        self.roi.remove_metadata(force=force)
        if self.roi_intensity is not None:
            self.roi_intensity.remove_metadata(force=force)
        if self.roi_morphology is not None:
            self.roi_morphology.remove_metadata(force=force)

    def to_object(self, **kwargs):
        return self.copy()

    def promote(self, **kwargs):
        return self

    def update_separate_slices(self, x):
        self.roi.update_separate_slices(x=x)
        if self.roi_intensity is not None:
            self.roi_intensity.update_separate_slices(x=x)
        if self.roi_morphology is not None:
            self.roi_morphology.update_separate_slices(x=x)

    def get_slices(
            self,
            slice_number: None | int | list[int] = None,
            primary_mask_only: bool = False
    ) -> None | Self | list[Self]:

        mask_list = []
        return_list = True

        if slice_number is None:
            slice_number = list(range(self.roi.image_dimension[0]))
        elif isinstance(slice_number, int):
            return_list = False
            slice_number = [slice_number]

        for current_slice_id in slice_number:
            slice_mask = self.copy(drop_image=True)
            slice_mask.roi = self.roi.get_slices(slice_number=current_slice_id)
            if slice_mask.roi_intensity is not None and not primary_mask_only:
                slice_mask.roi_intensity = self.roi_intensity.get_slices(slice_number=current_slice_id)
            else:
                slice_mask.roi_intensity = None

            if slice_mask.roi_morphology is not None and not primary_mask_only:
                slice_mask.roi_morphology = self.roi_morphology.get_slices(slice_number=current_slice_id)
            else:
                slice_mask.roi_morphology = None

            if slice_mask.is_empty():
                continue

            mask_list += [slice_mask]

        if len(mask_list) == 0:
            return None
        elif return_list:
            return mask_list
        else:
            return mask_list[0]

    def copy(self, drop_image=False) -> Self:

        # Create new mask by copying the current mask.
        mask = copy.deepcopy(self)

        if drop_image:
            mask.roi.drop_image()
            if mask.roi_intensity is not None:
                mask.roi_intensity.drop_image()
            if mask.roi_morphology is not None:
                mask.roi_morphology.drop_image()

        # Creates a new copy of the roi
        return mask

    def is_empty(self) -> bool:
        if self.roi is None:
            return True

        return self.roi.is_empty()

    def is_empty_mask(self):
        return self.roi.is_empty_mask()

    def interpolate(
            self,
            image: None | GenericImage,
            settings: SettingsClass
    ):
        # Skip if image and/or mask is missing
        if self.is_empty():
            return

        if image is None or image.is_empty():
            self.roi.interpolate(settings=settings)
            if self.roi_intensity is not None:
                self.roi_intensity.interpolate(settings=settings)
            if self.roi_morphology is not None:
                self.roi_morphology.interpolate(settings=settings)
        else:
            self.register(image=image, settings=settings)

    def register(
            self,
            image: GenericImage,
            spline_order: None | int = None,
            anti_aliasing: None | bool = None,
            anti_aliasing_smoothing_beta: None | float = None,
            settings: None | SettingsClass = None
    ):
        if (spline_order is None or anti_aliasing is None or anti_aliasing is None) and settings is None:
            raise ValueError("None of the parameters for registration can be set.")

        # Check if there is any mask data to work with.
        if self.is_empty():
            return

        # Check if the mask is empty.
        empty_before_registration = self.is_empty_mask()

        if spline_order is None:
            spline_order = self.roi.get_interpolation_spline_order(settings=settings)
        if anti_aliasing is None:
            anti_aliasing = settings.img_interpolate.anti_aliasing
        if anti_aliasing_smoothing_beta is None:
            anti_aliasing_smoothing_beta = settings.img_interpolate.smoothing_beta

        self.roi.register(
            image=image,
            spline_order=spline_order,
            anti_aliasing=anti_aliasing,
            anti_aliasing_smoothing_beta=anti_aliasing_smoothing_beta,
            mode="constant"
        )

        # Warn if a previously non-empty mask is empty after registration. This may indicate issues with the frame of
        # reference, i.e. world coordinate systems that do not have the same definition.
        if self.is_empty_mask() and not empty_before_registration:
            warnings.warn(
                f"The {self.roi_name} mask is empty after registering it to its image. Please check that mask and "
                f"image use the same frame of reference.",
                UserWarning
            )

        if self.roi_intensity is not None:
            self.roi_intensity.register(
                image=image,
                spline_order=spline_order,
                anti_aliasing=anti_aliasing,
                anti_aliasing_smoothing_beta=anti_aliasing_smoothing_beta,
                mode="constant"
            )
        if self.roi_morphology is not None:
            self.roi_morphology.register(
                image=image,
                spline_order=spline_order,
                anti_aliasing=anti_aliasing,
                anti_aliasing_smoothing_beta=anti_aliasing_smoothing_beta,
                mode="constant"
            )

    def merge(self, masks: list[Self]) -> Self:
        """Merge masks"""

        roi_mask = np.zeros(self.roi.image_dimension, dtype=bool)
        roi_name = []
        for mask in masks:
            # Skip empty masks.
            if mask.roi.is_empty():
                continue

            roi_mask = np.logical_or(roi_mask, mask.roi.get_voxel_grid())
            roi_name += [mask.roi_name]

        self.roi.set_voxel_grid(voxel_grid=roi_mask)
        self.roi_name = " + ".join(roi_name)

        if self.roi_intensity is not None:
            roi_mask = np.zeros(self.roi_intensity.image_dimension, dtype=bool)
            for mask in masks:
                # Skip empty masks.
                if mask.roi_intensity.is_empty():
                    continue

                roi_mask = np.logical_or(roi_mask, mask.roi_intensity.get_voxel_grid())

            self.roi_intensity.set_voxel_grid(voxel_grid=roi_mask)

        if self.roi_morphology is not None:
            roi_mask = np.zeros(self.roi_morphology.image_dimension, dtype=bool)
            for mask in masks:
                # Skip empty masks.
                if mask.roi_morphology.is_empty():
                    continue

                roi_mask = np.logical_or(roi_mask, mask.roi_morphology.get_voxel_grid())

            self.roi_morphology.set_voxel_grid(voxel_grid=roi_mask)

        return self

    def split_mask(self) -> list[Self]:
        """Split mask into multiple masks."""
        import skimage.measure

        if self.is_empty():
            return [self]

        # Label regions
        roi_label_mask, n_regions = skimage.measure.label(self.roi.get_voxel_grid(), connectivity=2, return_num=True)

        if n_regions == 1:
            return [self]

        new_masks = []
        for ii in np.arange(start=0, stop=n_regions):
            roi_mask = roi_label_mask == ii + 1

            new_mask = self.copy(drop_image=True)
            new_mask.roi.set_voxel_grid(roi_mask)
            if self.roi_intensity is not None:
                new_mask.roi_intensity.set_voxel_grid(voxel_grid=np.logical_and(roi_mask, self.roi_intensity.get_voxel_grid()))
            if self.roi_morphology is not None:
                new_mask.roi_morphology.set_voxel_grid(voxel_grid=np.logical_and(roi_mask, self.roi_morphology.get_voxel_grid()))

            new_mask.roi_name += f"-{ii + 1}"

            new_masks += [new_mask]

        return new_masks

    def select_largest_slice(self):
        """Crops to the largest slice."""

        if self.is_empty():
            return

        # Find axial slice that contains the largest part of the mask.
        roi_size = np.sum(self.roi.get_voxel_grid(), axis=(1, 2))
        if np.all(roi_size == 0):
            return

        # Find the index of said slice
        largest_slice_index = np.argmax(roi_size)

        # Copy only largest slice.
        roi_mask = np.zeros(self.roi.image_dimension, dtype=bool)
        roi_mask[largest_slice_index, :, :] = self.roi.get_voxel_grid()[largest_slice_index, :, :]
        self.roi.set_voxel_grid(voxel_grid=roi_mask)

        if self.roi_intensity is not None:
            roi_mask = np.zeros(self.roi_intensity.image_dimension, dtype=bool)
            roi_mask[largest_slice_index, :, :] = self.roi_intensity.get_voxel_grid()[largest_slice_index, :, :]
            self.roi_intensity.set_voxel_grid(voxel_grid=roi_mask)

        if self.roi_morphology is not None:
            roi_mask = np.zeros(self.roi_morphology.image_dimension, dtype=bool)
            roi_mask[largest_slice_index, :, :] = self.roi_morphology.get_voxel_grid()[largest_slice_index, :, :]
            self.roi_morphology.set_voxel_grid(voxel_grid=roi_mask)

    def select_largest_region(self):
        """Crops to the largest region."""
        import skimage.measure

        if self.is_empty():
            return

        # Label regions
        roi_label_mask, n_regions = skimage.measure.label(self.roi.get_voxel_grid(), connectivity=2, return_num=True)

        # Determine size of regions
        roi_sizes = np.zeros(n_regions)
        for ii in np.arange(start=0, stop=n_regions):
            roi_sizes[ii] = np.sum(roi_label_mask == ii + 1)

        # Select largest region
        roi_mask = roi_label_mask == np.argmax(roi_sizes) + 1
        self.roi.set_voxel_grid(voxel_grid=roi_mask)

        if self.roi_intensity is not None:
            self.roi_intensity.set_voxel_grid(voxel_grid=np.logical_and(self.roi_intensity.get_voxel_grid(), roi_mask))
        if self.roi_morphology is not None:
            self.roi_morphology.set_voxel_grid(voxel_grid=np.logical_and(self.roi_morphology.get_voxel_grid(), roi_mask))

    def generate_masks(self):
        """"Generate roi intensity and morphology masks"""

        if self.roi is None:
            self.roi_intensity = None
            self.roi_morphology = None
        else:
            if self.roi_intensity is None:
                self.roi_intensity = self.roi.copy()
            if self.roi_morphology is None:
                self.roi_morphology = self.roi.copy()

    def decimate(self):
        """
        Decimates the roi mask.
        """
        if self.roi is not None:
            self.roi.decimate()
        if self.roi_intensity is not None:
            self.roi_intensity.decimate()
        if self.roi_morphology is not None:
            self.roi_morphology.decimate()

    def crop(
            self,
            ind_ext_z=None,
            ind_ext_y=None,
            ind_ext_x=None,
            xy_only=False,
            z_only=False):

        # Crop masks.
        if self.roi is not None:
            self.roi.crop(
                ind_ext_z=ind_ext_z,
                ind_ext_y=ind_ext_y,
                ind_ext_x=ind_ext_x,
                xy_only=xy_only,
                z_only=z_only)

        if self.roi_intensity is not None:
            self.roi_intensity.crop(
                ind_ext_z=ind_ext_z,
                ind_ext_y=ind_ext_y,
                ind_ext_x=ind_ext_x,
                xy_only=xy_only,
                z_only=z_only)

        if self.roi_morphology is not None:
            self.roi_morphology.crop(
                ind_ext_z=ind_ext_z,
                ind_ext_y=ind_ext_y,
                ind_ext_x=ind_ext_x,
                xy_only=xy_only,
                z_only=z_only)

    def crop_to_size(self, center, crop_size):
        """"Crops roi to a pre-defined size"""

        # Crop masks to size
        if self.roi is not None:
            self.roi.crop_to_size(center=center, crop_size=crop_size)
        if self.roi_intensity is not None:
            self.roi_intensity.crop_to_size(center=center, crop_size=crop_size)
        if self.roi_morphology is not None:
            self.roi_morphology.crop_to_size(center=center, crop_size=crop_size)

    def resegmentise_mask(
            self,
            image: GenericImage,
            resegmentation_method: None | str | list[str] = None,
            intensity_range: None | tuple[Any, Any] = None,
            sigma: None | float = None
    ):
        # Resegmentation of the mask based on image intensities.

        # Set intensity range because this is used elsewhere.
        if intensity_range is not None:
            self.intensity_range = tuple(intensity_range)

        if image.is_empty() or self.is_empty():
            return

        # Ensure that masks are generated.
        self.generate_masks()

        if resegmentation_method is None or "none" in resegmentation_method:
            return

        # Initialise range.
        updated_range = [np.nan, np.nan]

        if any(method in ["threshold", "range"] for method in resegmentation_method):
            # Filter out voxels with intensity outside prescribed range

            if intensity_range is None:
                raise ValueError("Intensity range is not provided, but required for resegmentation.")

            # Upper threshold
            if not np.isnan(intensity_range[1]):
                updated_range[1] = copy.deepcopy(intensity_range[1])

            # Lower threshold
            if not np.isnan(intensity_range[0]):
                updated_range[0] = copy.deepcopy(intensity_range[0])

        if any(method in ["sigma", "outlier"] for method in resegmentation_method):
            # Remove voxels with outlier intensities

            # Local constant
            image_data = image.get_voxel_grid()
            mask_data = self.roi.get_voxel_grid()

            # Check if the voxel grid is not empty
            if np.any(mask_data):

                # Calculate mean and standard deviation of intensities in roi
                mean_int = np.mean(image_data[mask_data])
                sd_int = np.std(image_data[mask_data])

                if not np.isnan(updated_range[0]):
                    updated_range[0] = np.max([updated_range[0], mean_int - sigma * sd_int])
                else:
                    updated_range[0] = mean_int - sigma * sd_int

                if not np.isnan(updated_range[1]):
                    updated_range[1] = np.min([updated_range[1], mean_int + sigma * sd_int])
                else:
                    updated_range[1] = mean_int + sigma * sd_int

        if not np.isnan(updated_range[0]) or not np.isnan(updated_range[1]):
            # Update intensity mask
            intensity_mask_data = self.roi.get_voxel_grid()

            if not np.isnan(updated_range[0]):
                intensity_mask_data = np.logical_and((image.get_voxel_grid() >= updated_range[0]), intensity_mask_data)

            if not np.isnan(updated_range[1]):
                intensity_mask_data = np.logical_and((image.get_voxel_grid() <= updated_range[1]), intensity_mask_data)

            # Set roi voxel volume
            self.roi_intensity.set_voxel_grid(voxel_grid=intensity_mask_data)

    def dilate(
            self,
            distance: None | float = None,
            voxel_distance: None | float = None
    ):
        # Skip if the mask does not exist
        if self.roi is None:
            return

        self.roi.dilate(
            distance=distance,
            voxel_distance=voxel_distance
        )

    def erode(
            self,
            max_eroded_volume_fraction: float = 0.8,
            distance: None | float = None,
            voxel_distance: None | float = None
    ):
        # Skip if the mask does not exist
        if self.roi is None:
            return

        self.roi.erode(
            max_eroded_volume_fraction=max_eroded_volume_fraction,
            distance=distance,
            voxel_distance=voxel_distance
        )

    def fractional_volume_change(
            self,
            fractional_change: None | float = None
    ):
        # Skip if the mask does not exist
        if self.roi is None:
            return

        self.roi.fractional_volume_change(
            fractional_change=fractional_change
        )

    def decode_voxel_grid(self):
        """Converts run length encoded grids to conventional volumes"""

        # Decode main ROI object
        if self.roi is not None:
            self.roi.decode_voxel_grid()

        # Decode intensity and morphological masks
        if self.roi_intensity is not None:
            self.roi_intensity.decode_voxel_grid()
        if self.roi_morphology is not None:
            self.roi_morphology.decode_voxel_grid()

    def as_pandas_dataframe(
            self,
            image: None | GenericImage,
            intensity_mask: bool = False,
            morphology_mask: bool = False,
            distance_map: bool = False,
            by_slice: bool = False
    ) -> None | pd.DataFrame:

        # Check that the image and mask are present.
        if image.is_empty() or self.is_empty():
            return None

        # Check if the masks exist and assign if not.
        self.generate_masks()

        # Create table from test object
        img_dims = image.image_dimension
        index_id = np.arange(start=0, stop=np.prod(img_dims))
        coordinates = np.unravel_index(indices=index_id, shape=img_dims)
        df_img = pd.DataFrame({
            "index_id": index_id,
            "g": np.ravel(image.get_voxel_grid()),
            "x": coordinates[2],
            "y": coordinates[1],
            "z": coordinates[0]
        })

        if intensity_mask:
            df_img["roi_int_mask"] = np.ravel(self.roi_intensity.get_voxel_grid()).astype(bool)
        if morphology_mask:
            df_img["roi_morph_mask"] = np.ravel(self.roi_morphology.get_voxel_grid()).astype(bool)

        if distance_map:
            # Calculate distance by sequential border erosion
            from scipy.ndimage import generate_binary_structure, binary_erosion

            # Set up distance map and morphological voxel grid
            dist_map = np.zeros(img_dims)
            morph_voxel_grid = self.roi_morphology.get_voxel_grid()

            if by_slice:
                # Distances are determined in 2D
                binary_struct = generate_binary_structure(rank=2, connectivity=1)

                # Iterate over slices
                for ii in np.arange(0, img_dims[0]):
                    # Calculate distance by sequential border erosion
                    roi_eroded = morph_voxel_grid[ii, :, :]

                    # Iterate distance from border
                    while np.sum(roi_eroded) > 0:
                        roi_eroded = binary_erosion(roi_eroded, structure=binary_struct)
                        dist_map[ii, :, :] += roi_eroded * 1

            else:
                # Distances are determined in 3D
                binary_struct = generate_binary_structure(rank=3, connectivity=1)

                # Copy of roi morphology mask
                roi_eroded = copy.deepcopy(morph_voxel_grid)

                # Incrementally erode the morphological mask
                while np.sum(roi_eroded) > 0:
                    roi_eroded = binary_erosion(roi_eroded, structure=binary_struct)
                    dist_map += roi_eroded * 1

            # Update distance from border, as minimum distance is 1
            dist_map[morph_voxel_grid] += 1

            # Add distance map to table
            df_img["border_distance"] = np.ravel(dist_map).astype(int)

        return df_img

    def get_bounding_box(self):
        return self.roi.get_bounding_box()

    def compute_diagnostic_features(
            self,
            image: GenericImage,
            settings: SettingsClass,
            append_str: str | None = None
    ) -> pd.DataFrame:
        """ Creates diagnostic features for the ROI """

        # Set feature names
        feature_names = [
            "int_map_dim_x", "int_map_dim_y", "int_map_dim_z", "int_bb_dim_x", "int_bb_dim_y", "int_bb_dim_z",
            "int_vox_dim_x", "int_vox_dim_y", "int_vox_dim_z", "int_vox_count", "int_mean_int", "int_min_int",
            "int_max_int",
            "mrp_map_dim_x", "mrp_map_dim_y", "mrp_map_dim_z", "mrp_bb_dim_x", "mrp_bb_dim_y", "mrp_bb_dim_z",
            "mrp_vox_dim_x", "mrp_vox_dim_y", "mrp_vox_dim_z", "mrp_vox_count", "mrp_mean_int", "mrp_min_int",
            "mrp_max_int"
        ]

        # Create pandas dataframe with one row and feature columns
        df = pd.DataFrame(np.full(shape=(1, len(feature_names)), fill_value=np.nan))
        df.columns = feature_names

        # Skip further analysis if the image and/or roi are missing
        if image is None or image.is_empty() or self.is_empty():
            return df

        # Register with image on function call to ensure that mask and image correspond to the same space.
        mask_copy = self.copy()
        mask_copy.register(
            image=image,
            settings=settings
        )

        # Make copies of intensity and morphological masks (if required)
        mask_copy.generate_masks()

        # Get image and roi voxel grids
        img_voxel_grid = image.get_voxel_grid()
        int_voxel_grid = mask_copy.roi_intensity.get_voxel_grid()
        mrp_voxel_grid = mask_copy.roi_morphology.get_voxel_grid()

        # Compute bounding boxes
        int_bounding_box_dim = mask_copy.roi_intensity.get_bounding_box()
        mrp_bounding_box_dim = mask_copy.roi_morphology.get_bounding_box()
        if any(x is None for x in int_bounding_box_dim) or any(y is None for y in mrp_bounding_box_dim):
            return df

        int_bounding_box_dim = np.squeeze(np.diff(int_bounding_box_dim, axis=1) + 1)
        mrp_bounding_box_dim = np.squeeze(np.diff(mrp_bounding_box_dim, axis=1) + 1)

        # Set intensity mask features
        df["int_map_dim_x"] = mask_copy.roi_intensity.image_dimension[2]
        df["int_map_dim_y"] = mask_copy.roi_intensity.image_dimension[1]
        df["int_map_dim_z"] = mask_copy.roi_intensity.image_dimension[0]
        df["int_bb_dim_x"] = int_bounding_box_dim[2]
        df["int_bb_dim_y"] = int_bounding_box_dim[1]
        df["int_bb_dim_z"] = int_bounding_box_dim[0]
        df["int_vox_dim_x"] = mask_copy.roi_intensity.image_spacing[2]
        df["int_vox_dim_y"] = mask_copy.roi_intensity.image_spacing[1]
        df["int_vox_dim_z"] = mask_copy.roi_intensity.image_spacing[0]
        df["int_vox_count"] = np.sum(int_voxel_grid)
        df["int_mean_int"] = np.mean(img_voxel_grid[int_voxel_grid])
        df["int_min_int"] = np.min(img_voxel_grid[int_voxel_grid])
        df["int_max_int"] = np.max(img_voxel_grid[int_voxel_grid])

        # Set morphological mask features
        df["mrp_map_dim_x"] = mask_copy.roi_morphology.image_dimension[2]
        df["mrp_map_dim_y"] = mask_copy.roi_morphology.image_dimension[1]
        df["mrp_map_dim_z"] = mask_copy.roi_morphology.image_dimension[0]
        df["mrp_bb_dim_x"] = mrp_bounding_box_dim[2]
        df["mrp_bb_dim_y"] = mrp_bounding_box_dim[1]
        df["mrp_bb_dim_z"] = mrp_bounding_box_dim[0]
        df["mrp_vox_dim_x"] = mask_copy.roi_morphology.image_spacing[2]
        df["mrp_vox_dim_y"] = mask_copy.roi_morphology.image_spacing[1]
        df["mrp_vox_dim_z"] = mask_copy.roi_morphology.image_spacing[0]
        df["mrp_vox_count"] = np.sum(mrp_voxel_grid)
        df["mrp_mean_int"] = np.mean(img_voxel_grid[mrp_voxel_grid])
        df["mrp_min_int"] = np.min(img_voxel_grid[mrp_voxel_grid])
        df["mrp_max_int"] = np.max(img_voxel_grid[mrp_voxel_grid])

        # Update column names
        if append_str is None:
            df.columns = ["_".join(["diag", feature]) for feature in df.columns]
        else:
            df.columns = ["_".join(["diag", feature, append_str]) for feature in df.columns]

        del mask_copy

        return df

    def get_center_position(self) -> list[Any]:
        """Identify location of the geometric center of the roi."""
        # Return a NaN if no roi is present
        if self.roi is None:
            return [np.nan, np.nan, np.nan]

        # Determine indices of voxels included in the roi
        z_ind, y_ind, x_ind = np.where(self.roi.get_voxel_grid())

        return [np.mean(z_ind), np.mean(y_ind), np.mean(x_ind)]

    def get_center_slice(self):
        """Identify location of the central slice in the roi."""

        # Return a NaN if no roi is present
        if self.roi is None:
            return np.nan

        # Determine indices of voxels included in the roi
        z_ind, y_ind, x_ind = np.where(self.roi.get_voxel_grid())
        z_center = (np.max(z_ind) + np.min(z_ind)) // 2

        return z_center

    def write(
            self,
            dir_path: str | Path,
            write_all: bool = False,
            file_format: str = "nifti"
    ):
        """
        Write masks to file
        :param dir_path: Path to directory where the image should be written.
        :param write_all: If true, creates NIfTI files from both intensity and morphology (original) masks.
        :param file_format: File format for image file. Can be nifti or numpy.
        :return: Nothing.
        """

        roi_str_components = self.get_file_name_descriptor()

        if write_all:
            self.roi_morphology.write(
                dir_path=dir_path,
                file_name="_".join(roi_str_components + ["morph"]),
                file_format=file_format
            )
            self.roi_intensity.write(
                dir_path=dir_path,
                file_name="_".join(roi_str_components + ["int"]),
                file_format=file_format
            )
        else:
            self.roi.write(
                dir_path=dir_path,
                file_name="_".join(roi_str_components),
                file_format=file_format
            )

    def get_file_name_descriptor(self) -> list[str]:

        return self.roi.get_file_name_descriptor() + [self.roi_name]

    def export(
            self,
            write_all=False,
            export_format: str = "dict"
    ) -> np.ndarray | list[np.ndarray] | dict[str, Any] | Self:
        if self.is_empty():
            return None

        if export_format == "dict":
            attributes = self.get_export_attributes()

            if write_all:
                intensity_mask = None if self.roi_intensity is None else self.roi_intensity.get_voxel_grid()
                morphology_mask = None if self.roi_morphology is None else self.roi_morphology.get_voxel_grid()
                attributes.update({"intensity_mask": intensity_mask, "morphology_mask": morphology_mask})
            else:
                attributes.update({"mask": self.roi.get_voxel_grid()})

            return attributes

        elif export_format == "numpy":
            if write_all:
                intensity_mask = None if self.roi_intensity is None else self.roi_intensity.get_voxel_grid()
                morphology_mask = None if self.roi_morphology is None else self.roi_morphology.get_voxel_grid()
                return [intensity_mask, morphology_mask]
            else:
                return self.roi.get_voxel_grid()

        elif export_format == "native":
            return self.copy()

        else:
            raise ValueError(f"The current value of export_format was not recognised: {export_format}")

    def get_export_attributes(self) -> dict[str, Any]:
        attributes = dict([("roi_name", self.roi_name)])
        attributes.update(self.roi.get_export_attributes())

        return attributes

    def export_roi_labels(self):

        attributes = self.get_export_attributes()
        attributes["roi_label"] = attributes.pop("roi_name")

        # A list is expected for roi_label.
        attributes["roi_label"] = [attributes["roi_label"]]

        for remove_key in ["translation", "rotation", "image_spacing", "image_origin", "image_orientation"]:
            if remove_key in attributes.keys():
                attributes.pop(remove_key)

        return attributes